from functools import wraps

from fumus.utils import Optional, Result

# Naive implementations -> one could do better than this


def returns_optional(func):
    @wraps(func)
    def wrapper(*args, **kw):
        return Optional.of_nullable(func(*args, *kw))

    return wrapper


def returns_result(*err_list):
    def handle_errors(func):
        exceptions = tuple(set(err_list).union((Exception,)))

        @wraps(func)
        def invoke_func(*args, **kw):
            try:
                result = func(*args, *kw)
            except exceptions as err:
                return Result.failure(err)
            return Result.success(result)

        return invoke_func

    return handle_errors



from fumus.utils import Optional


class Result:
    __slots__ = (
        "value",
        "error",
    )

    def __init__(self, value=None, error=None):
        if value is None and error is None:
            raise ValueError("Result's Value and Error cannot both be None")
        self.value = value
        self.error = error

    @property
    def is_successful(self):
        """Returns bool whether the Result is successful and an error is not present"""
        return self.error is None

    @classmethod
    def success(cls, value):
        """Creates Result describing given value"""
        return cls(value)

    @classmethod
    def failure(cls, error):
        """Creates Result describing given error"""
        return cls(error=error)

    def map_success(self, mapper):
        """
        If a value is present, apply the provided mapping function to it,
        and if the result is non-null, return an Optional describing the result.
        Otherwise return an empty Optional.
        (NB: if the provided mapper returns an Optional,
        the result isn't wrapped-up in an additional one)
        """
        if self.is_successful:
            return self._map_result(mapper, self.value)
        return Optional.empty()

    def map_failure(self, mapper):
        """
        If an error is present, apply the provided mapping function to it,
        and if the result is non-null, return an Optional describing the result.
        Otherwise return an empty Optional
        """
        if not self.is_successful:
            return self._map_result(mapper, self.error)
        return Optional.empty()

    def map(self, on_success, on_failure):
        """Combines map_success() and map_failure()"""
        if self.is_successful:
            return self._map_result(on_success, self.value)
        return self._map_result(on_failure, self.error)

    def if_success(self, consumer):
        """Performs given action with the value if the Result is successful"""
        if self.is_successful:
            consumer(self.value)

    def if_failure(self, consumer):
        """Performs given action with the error if the Result is not successful"""
        if not self.is_successful:
            consumer(self.error)

    def handle(self, on_success, on_failure):
        """Combines if_success() and if_failure()"""
        if self.is_successful:
            on_success(self.value)
        else:
            on_failure(self.error)

    def or_else(self, other):
        """
        Returns the value if successful, or a provided argument otherwise
        """
        return self.value if self.is_successful else other

    def or_else_get(self, supplier):
        """
        Returns the value if present, or calls a 'supplier' function otherwise
        """
        return self.value if self.is_successful else supplier()

    def or_else_raise(self, supplier=None):
        """
        Returns the value if successful,
        otherwise throws an exception produced by the exception supplying function
        (if such is provided by the user) or re-raises original Exception
        """
        if self.is_successful:
            return self.value
        if supplier:
            supplier(self.error)
        raise self.error

    def __str__(self):
        return f"Result[value={self.value}, error={self.error}]"

    def __eq__(self, other):
        if self.is_successful and other.is_successful:
            return self.value == other.value
        return False

    def __hash__(self):
        return hash(self.value) if self.is_successful else 0

    # ### helpers ###
    @staticmethod
    def _map_result(mapper, arg):
        result = mapper(arg)
        if isinstance(result, Optional):
            return result
        return Optional.of_nullable(result)



from fumus.exceptions.exception import NoSuchElementError, NoneTypeError


class Optional:
    """Container object which may (or may not) contain a non-null value"""

    __slots__ = ("_element",)

    def __init__(self, element):
        self._element = element

    @classmethod
    def empty(cls):
        """Creates empty Optional"""
        return cls(None)

    @classmethod
    def of(cls, element):
        """Creates Optional describing given non-null value"""
        if element is None:
            raise NoneTypeError("Value cannot be None")
        return cls(element)

    @classmethod
    def of_nullable(cls, element):
        """
        Returns an Optional describing the given value, if non-null,
        otherwise returns an empty Optional
        """
        return cls(element)

    def get(self):
        """If a value is present, returns the value, otherwise raises an Exception"""
        if self.is_empty:
            raise NoSuchElementError("Optional is empty")
        return self._element

    @property
    def is_present(self):
        """Returns bool whether a value is present"""
        return not self.is_empty

    @property
    def is_empty(self):
        """Returns bool whether the Optional is empty"""
        return self._element is None

    def if_present(self, action):
        """Performs given action with the value if the Optional is not empty"""
        if self.is_present:
            action(self.get())

    def if_present_or_else(self, action, empty_action):
        """Performs given action with the value if the Optional is not empty,
        otherwise calls fallback 'empty_action'
        """
        if self.is_present:
            action(self.get())
        else:
            empty_action()

    def or_else(self, value):
        """
        Returns the value if present, or a provided argument otherwise.
        Safe alternative to get() method
        """
        return self._element if self.is_present else value

    def or_else_get(self, supplier):
        """
        Returns the value if present, or calls a 'supplier' function otherwise.
        Safe alternative to get() method
        """
        return self._element if self.is_present else supplier()

    def or_else_raise(self, supplier=None):
        """
        Returns the value if present,
        otherwise throws an exception produced by the exception supplying function
        (if such is provided by the user) or NoSuchElementError
        """
        if self.is_present:
            return self._element
        if supplier:
            supplier()
        raise NoSuchElementError("Optional is empty")

    def map(self, mapper):
        """
        If a value is present, apply the provided mapping function to it,
        and if the result is non-null, return an Optional describing the result.
        Otherwise return an empty Optional.
        (NB: if the provided mapper returns an Optional,
        the result isn't wrapped-up in an additional one)
        """
        if self.is_present:
            result = mapper(self._element)
            if isinstance(result, Optional):
                return result
            return Optional.of_nullable(result)
        return self

    def filter(self, predicate):
        """
        If a value is present, and the value matches the given predicate,
        returns an Optional describing the value, otherwise returns an empty Optional
        """
        if self.is_present:
            if predicate(self._element):
                return self
            return Optional.empty()
        return self

    def __repr__(self):
        if self.is_present:
            return f"Optional[{self._element}]"
        return "Optional.empty"

    def __eq__(self, other):
        if self.is_present and other.is_present:
            return self.get() == other.get()
        return self.is_empty and other.is_empty

    def __hash__(self):
        return hash(self._element) if self.is_present else 0
